-- TODO: Handle colors
local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local L = LibStub("AceLocale-3.0"):GetLocale("RW-2.0")

local Boss = RW.bossPrototype
local timerProt = {}
local mt = {__index = timerProt}

local Party = RW:GetPlugin("Party")
local BITS = RW.BITS

Boss.BITS = Boss.BITS or {}
Boss.BITS.BAR = {
	CD = 0x00000001,
	DUR = 0x00000002,
	CAST = 0x00000004,
	NORMAL = 0x00000008,
	DUR_TARGET = 0x00000010,
	DUR_TARGET_STACK = 0x00000020,
	DUR_STACK = 0x00000040,
	BERSERK = 0x00000080,
}

-- If text is a number this warning is considered not a custom one
-- If text is not a number then it is custom and it can be possibly sent as a table,
-- then the first value in the table will be the actual text, and the second oen will be the option text as its show in the GUI
-- 3rd value in the table will be the tooltip text
local custText = true
function Boss:_Timer(tType, time, text, icon, optionName, default, color, endMessage)
	if type(text) == "number" then 
		custText = false
		return self:_Timer(tType, time, GetSpellInfo(text), icon or text, optionName or text, default, color, endMessage) 	
	end
	if type(icon) == "number" then return self:_Timer(tType, time, text, select(3, GetSpellInfo(icon)), optionName, default, color, endMessage) end
	
	-- If option name is a number, then add it as a spell option and update it to the actual name
	-- Else add a gereral option
	if type(optionName) == "number" then
		if tType == "BERSERK" then
			self:AddGeneralOption(RW.SpellName[optionName])
		else
			self:AddSpellOption("General", optionName)
		end
		optionName = RW.SpellName[optionName]
	else
		self:AddGeneralOption(optionName or text)
	end
	-- If no option name is found by now we use the text as default
	if not optionName then optionName = type(text) == "table" and text[1] or text end
	
	-- Extract the info from the text table for custom option
	local customOptName
	if type(text) == "table" then
		customOptName = text[2]
		text = text[1]
	elseif custText then
		customOptName = text
	end
	
	-- Do the magic to add all the bits and stuff
	self:AddDefaults(optionName, "BAR", tType, default, customOptName)
	
	local defaultOption = self.defaultOptions[optionName]
	defaultOption.endMessage = endMessage
	defaultOption.Colors = defaultOption.Colors or {}
	defaultOption.Colors[customOptName or "BAR"..optionName] = color
	
	local obj =  setmetatable({
		type = tType,
		time = time,
		text = text,
		customText = custText,
		customOptName = customOptName,
		icon = icon,
		mod = self,
		id = optionName,
		running = {},
		scheduled = {},
		default = default
	}, mt)
	
	self.timers = self.timers or {}
	tinsert(self.timers, obj)
	
	custText = true
	
	return obj
end

function Boss:Timer(...)
	return self:_Timer("NORMAL", ...)
end

function Boss:CDTimer(...)
	return self:_Timer("CD", ...)
end

function Boss:DurTimer(...)
	return self:_Timer("DUR", ...)
end

function Boss:DurTargetTimer(...)
	return self:_Timer("DUR_TARGET", ...)
end

function Boss:DurTargetStackTimer(...)
	return self:_Timer("DUR_TARGET_STACK", ...)
end

function Boss:DurStackTimer(...)
	return self:_Timer("DUR_STACK", ...)
end

function Boss:BerserkTimer(time, text, ...)
	return self:_Timer("BERSERK", time, text or 35595, ...)
end

function Boss:CustomTimer(...)
	return self:_Timer("CUSTOM", ...)
end

function Boss:CastTimer(time, ...)
	if time > 1000 then
		local castTime = select(7, GetSpellInfo(time))
		local spellHaste = select(7, GetSpellInfo(53142)) / 10000
		local spell = time
		time = castTime/spellHaste/1000
		return self:CastTimer(time, spell, ...)
	end
	return self:_Timer("CAST", time, ...)
end

function Boss:StopAllTimers()
	if not self.timers then return end
	for i, timer in pairs(self.timers) do
		for j, runningID in pairs(timer.running) do
			RW:SendMessage("RW_CancelBar", runningID)
			timer.running[j] = nil
		end
	end
end

function Boss:RW_BarFinished(message, id, text, icon)
	if not self.timers then return end
	for i, timer in pairs(self.timers) do
		for j, runningID in pairs(timer.running) do
			if runningID == id then
				if self.db.profile[timer.id].endMessage then
					local c = timer.mod.db.profile[timer.id].Colors[timer.customOptName or "BAR"..timer.id]
					if not c then c = {1, 1, 1, 1} end
					self:SendMessage("RW_Message", text, c[1], c[2], c[3], c[4], icon)
				end
				timer.running[j] = nil
			end
		end
	end
end

function timerProt:SetIcon(icon)
	self.icon = icon
end

function timerProt:Start(time, info, ...)
	if bit.band(self.mod.db.profile[self.id].uses, RW.BITS.BAR) ~= RW.BITS.BAR then return end
	if self.customOptName then
		if not self.mod.db.profile[self.id]["BAR"..self.customOptName] then return end
	elseif bit.band(self.mod.db.profile[self.id].BAR, self.mod.BITS.BAR[self.type]) ~= self.mod.BITS.BAR[self.type] then return end
	if time and type(time) ~= "number" then return self:Start(nil, time, info, ...) end
	
	local text
	if self.customText then		
		text = RW:Format(self.text, info, ...)
	else
		text = RW:Format(L["FORMAT_BAR_"..self.type], self.text, info, ...)	
	end
	
	text = text:gsub(">([%S\s]+)<", function(name)
		if Party.group[name] then
			local color = RAID_CLASS_COLORS[Party.group[name].class] or self.color
			name = ("|cff%.2x%.2x%.2x%s|r"):format(color.r * 255, color.g * 255, color.b * 255, name)
		end
		return name
	end)
	
	local c = self.mod.db.profile[self.id].Colors[self.customOptName or "BAR"..self.id]
	RW:SendMessage("RW_StartBar", self.type..self.id..(info or ""), time or self.time, text, self.icon, c)
	tinsert(self.running, self.type..self.id..(info or ""))
end

function timerProt:Stop(extraID)
	if extraID then
		RW:SendMessage("RW_CancelBar", self.id..extraID)
	else
		for i, id in pairs(self.running) do
			RW:SendMessage("RW_CancelBar", id)
		end
	end
end
timerProt.Cancel = timerProt.Stop

local function startTimer(timer, ...)
	timer:Start(...)
end

function timerProt:Schedule(stime, time, info, ...)
	if time and type(time) ~= "number" then return self:Schedule(stime, nil, time, info, ...) end
	self.mod:Schedule("BAR"..self.id..(info or ""), stime, startTimer, self, time, info, ...)
	tinsert(self.scheduled, "BAR"..self.id..(info or ""))
end

function timerProt:ScheduleRepeating(stime, now, time, info, ...)
	if time and type(time) ~= "number" then return self:ScheduleRepeating(stime, now, nil, time, info, ...) end
	self.mod:ScheduleRepeating("BAR"..self.id..(info or ""), stime, startTimer, now, self, time, info, ...)
	tinsert(self.scheduled, "BAR"..self.id..(info or ""))
end

function timerProt:Unschedule(extraID)
	if not extraID then
		for i, scheduleID in pairs(self.scheduled) do
			self.mod:Unschedule(scheduleID)
			self.scheduled[i] = nil
		end
	else
		self.mod:Unschedule("BAR"..self.id..(extraID or ""))
		RW:TRemove(self.scheduled, "BAR"..self.id..(extraID or ""))
	end
end

-- Options
Boss.optionGets = Boss.optionGets or {}
function Boss.optionGets.BAR(self, option)
	local AceGUI = LibStub("AceGUI-3.0")
	local group = AceGUI:Create("InlineGroup")
	group:SetFullWidth(true)
	group:SetTitle("Options")
	
	for i, timer in pairs(self.timers) do
		if timer.id == option.text then
			local subGroup = AceGUI:Create("SimpleGroup")
			subGroup:SetFullWidth(true)
			subGroup:SetLayout("flow")

			local check = AceGUI:Create("CheckBox")
			if timer.customOptName then
				check:SetLabel(timer.customOptName)
				check:SetValue(self.db.profile[option.text]["BAR"..timer.customOptName])
				check:SetUserData("optName", "BAR"..timer.customOptName)
				check:SetCallback("OnValueChanged", self.optionHelpers.onCustomValueChanged)
			else
				check:SetLabel(L["BAR_"..timer.type]) -- TODO: Set more info in the text, (the actual time for the bar)
				check:SetValue(bit.band(self.db.profile[option.text].BAR, Boss.BITS.BAR[timer.type]) == Boss.BITS.BAR[timer.type])
				check:SetUserData("type", "BAR")
				check:SetUserData("subType", timer.type)
				check:SetCallback("OnValueChanged", self.optionHelpers.onValueChanged)
				check:SetCallback("OnEnter", self.optionHelpers.onOptionEnter)
			end			
			check:SetUserData("option", option)
			check:SetUserData("boss", self)
			check:SetCallback("OnLeave", self.optionHelpers.onOptionLeave)
			
			local color = AceGUI:Create("ColorPicker")
			local colorvalue = self.db.profile[option.text].Colors[timer.customOptName or "BAR"..timer.id]
			if colorvalue then 
				color:SetColor(unpack(colorvalue))
			end
			color:SetLabel("Color")
			color:SetUserData("option", option)
			color:SetUserData("boss", self)
			color:SetUserData("key", "Colors")
			color:SetUserData("subKey", timer.customOptName or "BAR"..timer.id)
			color:SetCallback("OnValueConfirmed", self.optionHelpers.onValueChanged2)
			
			subGroup:AddChild(check)
			subGroup:AddChild(color)
			
			group:AddChild(subGroup)
		end
	end
	return group
end 